from datetime import datetime
import logging
import threading

from krbticket.command import KrbCommand
from krbticket.config import KrbConfig
from krbticket.updater import KrbTicketUpdater

logger = logging.getLogger(__name__)


class NoCredentialFound(Exception):
    pass


class KrbTicket():

    __instances__ = {}
    __instances_lock__ = threading.Lock()

    def __init__(self, config=None, file=None, principal=None, starting=None, expires=None,
                 service_principal=None, renew_expires=None):

        self.config = config
        self.file = file
        self.principal = principal
        self.starting = starting
        self.expires = expires
        self.service_principal = service_principal
        self.renew_expires = renew_expires
        self._updater = None
        self._updater_lock = threading.RLock()

    def updater_start(self, interval=KrbTicketUpdater.DEFAULT_INTERVAL):
        with self._updater_lock:
            updater = self.updater(interval=interval)
            updater.start()

    def updater(self, interval=KrbTicketUpdater.DEFAULT_INTERVAL):
        with self._updater_lock:
            if not self._updater:
                self._updater = self.config.updater_class(self, interval=interval)
        return self._updater

    def maybe_update(self):
        self.reload()

        if self.is_expired():
            if self.is_renewalable():
                self.renewal()
            else:
                self.reinit()

    def update(self, **kwargs):
        self.__dict__.update(**kwargs)

    def renewal(self):
        logger.info("Renewing ticket for {}...".format(self.principal))
        KrbCommand.renewal(self.config)
        self.reload()

    def reinit(self):
        logger.info("Reinitialize ticket for {}...".format(self.principal))
        KrbCommand.kinit(self.config)
        self.reload()

    def reload(self):
        logger.debug(
            "Reloading ticket attributes from {}...".format(self.file))
        KrbTicket.get_by_config(self.config)
        logger.debug(
            "Reloaded ticket attributes: {}...".format(self))

    def is_expired(self):
        return self.expires < self.config.renewal_threshold + datetime.now()

    def is_renewalable(self):
        if self.renew_expires:
            return self.renew_expires > self.config.renewal_threshold + datetime.now()
        else:
            return False

    def __str__(self):
        super_str = super(KrbTicket, self).__str__()
        return "{}: file={}, principal={}, starting={}, expires={}," \
               " service_principal={}, renew_expires={}" \
               .format(super_str, self.file, self.principal, self.starting,
                       self.expires, self.service_principal, self.renew_expires)

    @staticmethod
    def get_instance(**kwargs):
        ccache_name = kwargs.get('config').ccache_name
        with KrbTicket.__instances_lock__:
            if ccache_name not in KrbTicket.__instances__:
                KrbTicket.__instances__.update({ccache_name: KrbTicket(**kwargs)})
        ticket = KrbTicket.__instances__.get(ccache_name)
        ticket.update(**kwargs)
        return ticket

    @staticmethod
    def cache_exists(config):
        return KrbCommand.cache_exists(config)

    @staticmethod
    def init(principal, keytab=None, **kwargs):
        config = KrbConfig(principal=principal, keytab=keytab, **kwargs)
        return KrbTicket.init_by_config(config)

    @staticmethod
    def init_by_config(config):
        KrbCommand.kinit(config)
        return KrbTicket.get_by_config(config)

    @staticmethod
    def get_or_init(principal, keytab=None, **kwargs):
        config = KrbConfig(principal=principal, keytab=keytab, **kwargs)
        try:
            return KrbTicket.get_by_config(config)
        except NoCredentialFound:
            return KrbTicket.init_by_config(config)

    @staticmethod
    def get(principal, keytab=None, **kwargs):
        config = KrbConfig(principal=principal, keytab=keytab, **kwargs)
        return KrbTicket.get_by_config(config)

    @staticmethod
    def get_by_config(config):
        if KrbTicket.cache_exists(config):
            return KrbTicket.parse_from_klist(config, KrbCommand.klist(config))
        else:
            raise NoCredentialFound()

    @staticmethod
    def parse_from_klist(config, output):
        if not output:
            return KrbTicket.get_instance(config=config)

        lines = output.splitlines()
        file = lines[0].split(':')[2]
        principal = lines[1].split(':')[1].strip()
        starting, expires, service_principal = lines[4].strip().split('  ')
        if len(lines) > 5:
            renew_expires = lines[5].strip().replace('renew until ', '')
        else:
            renew_expires = None

        def parseDatetime(str):
            if str:
                return datetime.strptime(str, '%m/%d/%y %H:%M:%S')

        return KrbTicket.get_instance(
            config=config,
            file=file,
            principal=principal,
            starting=parseDatetime(starting),
            expires=parseDatetime(expires),
            service_principal=service_principal,
            renew_expires=parseDatetime(renew_expires))

    @staticmethod
    def _destroy():
        """
        destroy internal ticket registry

        stop all updaters belonging to a ticket registered in registry, and remove all entiries
        """
        with KrbTicket.__instances_lock__:
            for (key, ticket) in KrbTicket.__instances__.items():
                ticket.updater().stop()
                KrbCommand.kdestroy(ticket.config)
            KrbTicket.__instances__ = {}



from krbticket import KrbConfig
import os


def test_ccache_name():
    assert KrbConfig().ccache_name == '/tmp/krb5cc_{}'.format(os.getuid())
    assert KrbConfig(ccache_name="/tmp/hoge").ccache_name == '/tmp/hoge'
    os.environ['KRB5CCNAME'] = '/tmp/env_krb5cc'
    assert KrbConfig().ccache_name == '/tmp/env_krb5cc'
    del os.environ['KRB5CCNAME']



from datetime import timedelta
import logging
import multiprocessing
import os

logger = logging.getLogger(__name__)


class KrbConfig():
    from krbticket.updater import SimpleKrbTicketUpdater

    def __init__(self, principal=None, keytab=None, kinit_bin="kinit",
                 klist_bin="klist", kdestroy_bin="kdestroy",
                 renewal_threshold=timedelta(minutes=30),
                 ticket_lifetime=None,
                 ticket_renewable_lifetime=None,
                 ccache_name=None,
                 updater_class=SimpleKrbTicketUpdater,
                 retry_options={
                     'wait_exponential_multiplier': 1000,
                     'wait_exponential_max': 30000,
                     'stop_max_attempt_number': 10}):
        self.principal = principal
        self.keytab = keytab
        self.kinit_bin = kinit_bin
        self.klist_bin = klist_bin
        self.kdestroy_bin = kdestroy_bin
        self.renewal_threshold = renewal_threshold
        self.ticket_lifetime = ticket_lifetime
        self.ticket_renewable_lifetime = ticket_renewable_lifetime
        self.updater_class = updater_class
        self.retry_options = retry_options
        self.ccache_name = ccache_name if ccache_name else self._ccache_name()
        self.ccache_lockfile = '{}.krbticket.lock'.format(self.ccache_name)
        self.ccache_cmd_lockfile = '{}.krbticket.cmd.lock'.format(self.ccache_name)

    def __str__(self):
        super_str = super(KrbConfig, self).__str__()
        return "{}: principal={}, keytab={}, kinit_bin={}," \
               " klist_bin={}, kdestroy_bin={}, " \
               " renewal_threshold={}, ticket_lifetime={}, " \
               " ticket_renewable_lifetime={}, " \
               " retry_options={}, ccache_name={}, " \
               " updater_class={}" \
               .format(super_str, self.principal, self.keytab, self.kinit_bin,
                       self.klist_bin, self.kdestroy_bin,
                       self.renewal_threshold, self.ticket_lifetime,
                       self.ticket_renewable_lifetime,
                       self.retry_options, self.ccache_name,
                       self.updater_class)

    def _ccache_name(self):
        if self.updater_class.use_per_process_ccache():
            return self._per_process_ccache_name()
        else:
            return self._default_ccache_name()

    def _is_main_process(self):
        return multiprocessing.current_process().name == 'MainProcess'

    def _default_ccache_name(self):
        return os.environ.get('KRB5CCNAME', '/tmp/krb5cc_{}'.format(os.getuid()))

    def _per_process_ccache_name(self):
        if self._is_main_process():
            return self._default_ccache_name()

        new_ccname = "/tmp/krb5cc_{}_{}".format(os.getuid(), os.getpid())
        # Update KRB5CCNAME for kinit
        os.environ['KRB5CCNAME'] = new_ccname
        logger.info("env KRB5CCNAME is updated to '{}' for multiprocessing".format(new_ccname))

        return os.environ.get('KRB5CCNAME')



from krbticket import KrbTicket, KrbCommand, NoCredentialFound
from helper import *
from datetime import datetime
from freezegun import freeze_time
import os
import subprocess

def teardown_function(function):
    KrbTicket._destroy()


def test_init(config):
    KrbCommand.kdestroy(config)
    KrbTicket.init(DEFAULT_PRINCIPAL, DEFAULT_KEYTAB)


def test_get_or_init(config):
    KrbCommand.kdestroy(config)
    ticket0 = KrbTicket.get_or_init(DEFAULT_PRINCIPAL, DEFAULT_KEYTAB)
    ticket1 = KrbTicket.get_or_init(DEFAULT_PRINCIPAL, DEFAULT_KEYTAB)
    assert_ticket(ticket0, ticket1)


def test_init_with_keytab_env(config):
    KrbCommand.kdestroy(config)
    os.environ['KRB5_KTNAME'] = config.keytab
    config.keytab = None
    KrbTicket.init(DEFAULT_PRINCIPAL)
    del os.environ['KRB5_KTNAME']
    assert not os.environ.get('KRB5_KTNAME')


def test_init_without_keytab_env(config):
    KrbCommand.kdestroy(config)
    assert not os.environ.get('KRB5_KTNAME')
    retry_options = {
            'wait_exponential_multiplier': 100,
            'wait_exponential_max': 1000,
            'stop_max_attempt_number': 1 }
    config.keytab = None
    try:
        KrbTicket.init(DEFAULT_PRINCIPAL, retry_options=retry_options)
        pytest.fail()
    except subprocess.CalledProcessError:
        pass


def test_init_with_config(config):
    KrbCommand.kdestroy(config)
    ticket1 = KrbTicket.init_by_config(config)
    ticket2 = KrbTicket.init(
            DEFAULT_PRINCIPAL,
            DEFAULT_KEYTAB,
            renewal_threshold=timedelta(seconds=DEFAULT_TICKET_RENEWAL_THRESHOLD_SEC),
            ticket_lifetime=DEFAULT_TICKET_LIFETIME,
            ticket_renewable_lifetime=DEFAULT_TICKET_RENEWABLE_LIFETIME,
            retry_options={
                'wait_exponential_multiplier': 100,
                'wait_exponential_max': 1000,
                'stop_max_attempt_number': 3})
    assert_ticket(ticket1, ticket2)


def test_get(config):
    KrbCommand.kdestroy(config)
    with pytest.raises(NoCredentialFound):
        KrbTicket.get(DEFAULT_PRINCIPAL, DEFAULT_KEYTAB)

    assert_ticket(
        KrbTicket.init_by_config(config),
        KrbTicket.get(
            DEFAULT_PRINCIPAL,
            DEFAULT_KEYTAB,
            renewal_threshold=timedelta(seconds=DEFAULT_TICKET_RENEWAL_THRESHOLD_SEC),
            ticket_lifetime=DEFAULT_TICKET_LIFETIME,
            ticket_renewable_lifetime=DEFAULT_TICKET_RENEWABLE_LIFETIME,
            retry_options={
                'wait_exponential_multiplier': 100,
                'wait_exponential_max': 1000,
                'stop_max_attempt_number': 3}))

def test_object_uniqueness(config):
    KrbCommand.kdestroy(config)
    ticket0 = KrbTicket.init_by_config(config)
    ticket1 = KrbTicket.init_by_config(config)
    ticket2 = KrbTicket.get_by_config(config)

    config.ccache_name = '/tmp/krb5cc_{}'.format('dummy')
    ticket3 = KrbTicket.init_by_config(config)

    assert ticket0 == ticket1
    assert ticket0 == ticket2
    assert ticket0 != ticket3
    KrbCommand.kdestroy(config)

def test_ticket(config):
    KrbCommand.kdestroy(config)
    ticket = KrbTicket.init_by_config(config)
    assert_config(ticket.config, config)
    assert ticket.file
    assert ticket.principal == 'user@EXAMPLE.COM'
    assert ticket.starting
    assert ticket.expires
    assert ticket.service_principal
    assert ticket.renew_expires


def test_parse_klist_output(config):
    output = """
Ticket cache: FILE:/tmp/krb5cc_1000
Default principal: user@EXAMPLE.COM

Valid starting     Expires            Service principal
11/22/19 00:23:10  11/22/19 00:23:12  krbtgt/EXAMPLE.COM@EXAMPLE.COM
        renew until 12/20/19 00:23:10
""".strip()
    ticket = KrbTicket.parse_from_klist(config, output)
    assert ticket.principal == 'user@EXAMPLE.COM'
    assert ticket.service_principal == 'krbtgt/EXAMPLE.COM@EXAMPLE.COM'
    assert ticket.starting == datetime(2019, 11, 22, 0, 23, 10)
    assert ticket.expires == datetime(2019, 11, 22, 0, 23, 12)
    assert ticket.renew_expires == datetime(2019, 12, 20, 0, 23, 10)

@freeze_time("2019-11-20 00:00:00")
def test_expires(config):
    ticket = KrbTicket(
            config,
            starting = datetime(2019, 11, 10, 0, 0, 0),
            expires = datetime(2019, 11, 21, 0, 0, 0),
            renew_expires = datetime(2019, 11, 24, 0, 0, 0),
            )
    assert ticket.is_expired() == False;
    assert ticket.is_renewalable() == True;

    ticket = KrbTicket(
            config,
            starting = datetime(2019, 11, 10, 0, 0, 0),
            expires = datetime(2019, 11, 21, 0, 0, 0),
            renew_expires = None,
            )
    assert ticket.is_expired() == False;
    assert ticket.is_renewalable() == False;

    ticket = KrbTicket(
            config,
            starting = datetime(2019, 11, 10, 0, 0, 0),
            expires = datetime(2019, 11, 19, 0, 0, 0),
            renew_expires = datetime(2019, 11, 24, 0, 0, 0),
            )
    assert ticket.is_expired() == True;
    assert ticket.is_renewalable() == True;

    ticket = KrbTicket(
            config,
            starting = datetime(2019, 11, 10, 0, 0, 0),
            expires = datetime(2019, 11, 15, 0, 0, 0),
            renew_expires = datetime(2019, 11, 19, 0, 0, 0),
            )
    assert ticket.is_expired() == True;
    assert ticket.is_renewalable() == False;





import fasteners
import logging
import os
import subprocess
import threading
from retrying import retry

logger = logging.getLogger(__name__)
# fasteners is for interprocess multiprocessing
# threading.Lock is for multithreading
lock = threading.Lock()


class KrbCommand():
    @staticmethod
    def kinit(config):
        commands = []
        commands.append(config.kinit_bin)
        if config.ticket_lifetime:
            commands.append("-l")
            commands.append(config.ticket_lifetime)
        if config.ticket_renewable_lifetime:
            commands.append("-r")
            commands.append(config.ticket_renewable_lifetime)
        if config.ccache_name:
            commands.append("-c")
            commands.append(config.ccache_name)
        commands.append("-k")
        if config.keytab:
            commands.append("-t")
            commands.append(config.keytab)
        commands.append(config.principal)

        KrbCommand._call(config, commands)

    @staticmethod
    def renewal(config):
        commands = []
        commands.append(config.kinit_bin)
        if config.ccache_name:
            commands.append("-c")
            commands.append(config.ccache_name)
        commands.append("-R")
        commands.append(config.principal)

        KrbCommand._call(config, commands)

    @staticmethod
    def klist(config):
        commands = []
        commands.append(config.klist_bin)
        if config.ccache_name:
            commands.append("-c")
            commands.append(config.ccache_name)
        return KrbCommand._call(config, commands)

    @staticmethod
    def kdestroy(config):
        commands = []
        commands.append(config.kdestroy_bin)
        if config.ccache_name:
            commands.append("-c")
            commands.append(config.ccache_name)
        return KrbCommand._call(config, commands)

    @staticmethod
    def cache_exists(config):
        with lock:
            with fasteners.InterProcessLock(config.ccache_cmd_lockfile):
                return os.path.isfile(config.ccache_name)

    @staticmethod
    def _call(config, commands):

        def error_on_retry(exception):
            # will not retry if command is not found.
            if type(exception) == FileNotFoundError:
                raise exception

            logger.warning("the command failed. attempting retry... retry_options={}".format(config.retry_options))
            return True

        retry_options = {**config.retry_options, **{'retry_on_exception': error_on_retry}}
        @retry(**retry_options)
        def retriable_call():
            logger.debug("Executing {}".format(" ".join(commands)))
            custom_env = os.environ.copy()
            custom_env["LC_ALL"] = "C"
            return subprocess.check_output(commands, universal_newlines=True, env=custom_env)

        with lock:
            with fasteners.InterProcessLock(config.ccache_cmd_lockfile):
                return retriable_call()



from krbticket import KrbConfig, KrbCommand
from krbticket import MultiProcessKrbTicketUpdater
from helper import *
import os
import subprocess
from multiprocessing import Process
import pytest


def _test_commands(config):
    KrbCommand.kdestroy(config)
    KrbCommand.kinit(config)
    KrbCommand.renewal(config)
    KrbCommand.klist(config)
    KrbCommand.kdestroy(config)


def test_commands(config):
    _test_commands(config)


def test_commands_with_keytab_env(config):
    os.environ['KRB5_KTNAME'] = config.keytab
    config.keytab = None
    _test_commands(config)
    del os.environ['KRB5_KTNAME']
    assert not os.environ.get('KRB5_KTNAME')


def test_commands_without_keytab_env(config):
    assert not os.environ.get('KRB5_KTNAME')
    config.retry_options={
            'wait_exponential_multiplier': 100,
            'wait_exponential_max': 1000,
            'stop_max_attempt_number': 1}
    config.keytab = None
    try:
        _test_commands(config)
        pytest.fail()
    except subprocess.CalledProcessError:
        pass


def test_multiprocessing_with_per_process_ccache(config):
    KrbCommand.kdestroy(config)

    def run():
        conf = default_config()
        assert conf.ccache_name != config.ccache_name
        assert os.environ.get('KRB5CCNAME') == conf.ccache_name
        # check if KRB5CCNAME is recognized kerberos commands
        conf.ccache_name = None
        _test_commands(conf)

    processes = [Process(target=run) for i in range(10)]
    for p in processes:
        p.start()

    for p in processes:
        p.join()
        assert not p.exitcode


def test_multiprocessing_without_per_process_ccache():
    config = default_config(updater_class=MultiProcessKrbTicketUpdater)
    KrbCommand.kdestroy(config)

    def run():
        conf = default_config(updater_class=MultiProcessKrbTicketUpdater)
        assert conf.ccache_name == config.ccache_name
        assert not os.environ.get('KRB5CCNAME')

    processes = [Process(target=run) for i in range(10)]
    for p in processes:
        p.start()

    for p in processes:
        p.join()
        assert not p.exitcode


def test_retry(config, mocker):
    KrbCommand.kdestroy(config)

    raise_exception_twice = [
        subprocess.CalledProcessError(1, ['kinit']),
        subprocess.CalledProcessError(1, ['kinit']),
        None]
    patcher = mocker.patch('subprocess.check_output', side_effect=raise_exception_twice)
    KrbCommand.kinit(config)
    assert patcher.call_count == 3


def test_no_retry_when_filenotfound(config, mocker):
    KrbCommand.kdestroy(config)

    raise_exception = [
        FileNotFoundError,
        None]
    patcher = mocker.patch('subprocess.check_output', side_effect=raise_exception)
    try:
        KrbCommand.kinit(config)
        pytest.fail()
    except FileNotFoundError:
        assert patcher.call_count == 1


@pytest.mark.parametrize('config,expected', [
    (KrbConfig(
        principal=DEFAULT_PRINCIPAL,
        keytab=DEFAULT_KEYTAB,
        ticket_lifetime=DEFAULT_TICKET_LIFETIME,
        ticket_renewable_lifetime=DEFAULT_TICKET_RENEWABLE_LIFETIME),
    ['kinit', '-l', DEFAULT_TICKET_LIFETIME, '-r', DEFAULT_TICKET_RENEWABLE_LIFETIME, '-c', DEFAULT_CCACHE_NAME, '-k', '-t', './tests/conf/krb5.keytab', 'user@EXAMPLE.COM']),
    (KrbConfig(
        principal='alice@EXAMPLE.COM',
        keytab=DEFAULT_KEYTAB,
        ticket_lifetime=DEFAULT_TICKET_LIFETIME,
        ticket_renewable_lifetime=DEFAULT_TICKET_RENEWABLE_LIFETIME),
    ['kinit', '-l', DEFAULT_TICKET_LIFETIME, '-r', DEFAULT_TICKET_RENEWABLE_LIFETIME, '-c', DEFAULT_CCACHE_NAME, '-k', '-t', './tests/conf/krb5.keytab', 'alice@EXAMPLE.COM']),
    (KrbConfig(
        principal=DEFAULT_PRINCIPAL,
        keytab=None,
        ticket_lifetime='1s',
        ticket_renewable_lifetime=DEFAULT_TICKET_RENEWABLE_LIFETIME),
    ['kinit', '-l', '1s', '-r', DEFAULT_TICKET_RENEWABLE_LIFETIME, '-c', DEFAULT_CCACHE_NAME, '-k', 'user@EXAMPLE.COM']),
    (KrbConfig(
        principal=DEFAULT_PRINCIPAL,
        keytab=DEFAULT_KEYTAB,
        ticket_lifetime=DEFAULT_TICKET_LIFETIME,
        ticket_renewable_lifetime='6s'),
    ['kinit', '-l', DEFAULT_TICKET_LIFETIME, '-r', '6s', '-c', DEFAULT_CCACHE_NAME, '-k', '-t', './tests/conf/krb5.keytab', 'user@EXAMPLE.COM']),
    ])
def test_kinit_command(config, expected, mocker):
    mocker.patch.object(KrbCommand, '_call')
    KrbCommand.kinit(config)
    KrbCommand._call.assert_called_with(config, expected)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


def read_requirements(file):
    with open(file, "r") as f:
        return f.read().split()


install_requires = read_requirements("requirements.txt")
test_require = read_requirements("requirements-test.txt")
extras = {
    'test': test_require
}

setuptools.setup(
    name="krbticket",
    use_scm_version=True,
    author="Kazuhiro Suzuki",
    author_email="ksauzzmsg@gmail.com",
    description="Kerberos Ticket Manager",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ksauzz/krbticket",
    packages=setuptools.find_packages(),
    setup_requires=['setuptools_scm'],
    install_requires=install_requires,
    tests_require=test_require,
    extras_require=extras,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
    ],
)



from krbticket import KrbTicket, KrbCommand
from krbticket import SimpleKrbTicketUpdater, SingleProcessKrbTicketUpdater, MultiProcessKrbTicketUpdater
from helper import *
import time
import pytest
from multiprocessing import Process
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor


def teardown_function(function):
    KrbTicket._destroy()


def test_updater(config):
    KrbCommand.kdestroy(config)
    ticket = KrbTicket.init_by_config(config)
    updater = ticket.updater(interval=1)
    updater.start()
    assert updater.is_alive()
    updater.stop()
    time.sleep(2)
    assert not updater.is_alive()


def test_reuse_updater(config):
    KrbCommand.kdestroy(config)
    ticket = KrbTicket.init_by_config(config)
    updater = ticket.updater(interval=1)
    updater.start()
    assert updater.is_alive()
    updater.stop()
    time.sleep(2)
    assert not updater.is_alive()
    updater.start()
    assert not updater.is_alive()


def test_updater_start(config):
    KrbCommand.kdestroy(config)
    ticket = KrbTicket.init_by_config(config)
    ticket.updater_start(interval=1)
    assert ticket.updater().is_alive()
    ticket.updater().stop()
    time.sleep(2)
    assert not ticket.updater().is_alive()

@pytest.mark.parametrize('config', [
    default_config(updater_class=SimpleKrbTicketUpdater),
    default_config(updater_class=MultiProcessKrbTicketUpdater),
    default_config(updater_class=SingleProcessKrbTicketUpdater)
])
def test_renewal(config):
    """
    This test assumes:
    - 1 sec renewal threshold
    - 4 sec ticket lifetime
    - 8 sed renewal ticket lifetime
    """
    KrbCommand.kdestroy(config)
    ticket = KrbTicket.init_by_config(config)

    starting = ticket.starting
    expires = ticket.expires
    renew_expires = ticket.renew_expires

    updater = ticket.updater(interval=0.5)
    updater.start()

    # expect ticket renewal
    time.sleep(DEFAULT_TICKET_LIFETIME_SEC + DEFAULT_TICKET_RENEWAL_THRESHOLD_SEC)
    assert ticket.starting > starting
    assert ticket.expires > expires
    assert ticket.renew_expires == renew_expires

    starting = ticket.starting
    expires = ticket.expires

    # expect ticket re-initialize
    time.sleep(DEFAULT_TICKET_RENEWABLE_LIFETIME_SEC - DEFAULT_TICKET_LIFETIME_SEC + DEFAULT_TICKET_RENEWAL_THRESHOLD_SEC)
    assert ticket.starting > starting
    assert ticket.expires > expires
    assert ticket.renew_expires > renew_expires
    updater.stop()

@pytest.mark.parametrize('config_str', [
    'default_config(updater_class=SimpleKrbTicketUpdater)',
    'default_config(updater_class=MultiProcessKrbTicketUpdater)',
    'default_config(updater_class=SingleProcessKrbTicketUpdater)'
])
def test_multiprocessing_renewal(config_str, caplog):
    KrbCommand.kdestroy(eval(config_str))
    ticket = KrbTicket.init_by_config(eval(config_str))

    def run():
        """
        This test assumes:
        - 1 sec renewal threshold
        - 4 sec ticket lifetime
        - 8 sed renewal ticket lifetime
        """

        starting = ticket.starting
        expires = ticket.expires
        renew_expires = ticket.renew_expires

        updater = ticket.updater(interval=0.5)
        updater.start()

        # expect ticket renewal
        time.sleep(DEFAULT_TICKET_LIFETIME_SEC + DEFAULT_TICKET_RENEWAL_THRESHOLD_SEC)
        if 'SingleProcessKrbTicketUpdater' in config_str:
            # needs manual reload since SingleProcessTicketUpdater is stopped
            ticket.reload()
        assert ticket.starting > starting
        assert ticket.expires > expires
        assert ticket.renew_expires == renew_expires

        starting = ticket.starting
        expires = ticket.expires

        # expect ticket re-initialize
        time.sleep(DEFAULT_TICKET_RENEWABLE_LIFETIME_SEC - DEFAULT_TICKET_LIFETIME_SEC + DEFAULT_TICKET_RENEWAL_THRESHOLD_SEC)
        if 'SingleProcessKrbTicketUpdater' in config_str:
            # needs manual reload since SingleProcessTicketUpdater is stopped
            ticket.reload()
        assert ticket.starting > starting
        assert ticket.expires > expires
        assert ticket.renew_expires > renew_expires
        updater.stop()

    processes = [Process(target=run) for i in range(10)]
    for p in processes:
        p.start()

    for p in processes:
        p.join()
        assert not p.exitcode


@pytest.mark.parametrize('config_str', [
    'default_config(updater_class=SimpleKrbTicketUpdater)',
    'default_config(updater_class=MultiProcessKrbTicketUpdater)',
    'default_config(updater_class=SingleProcessKrbTicketUpdater)'
])
def test_single_thread_updater_in_multithreading(config_str, caplog):
    KrbCommand.kdestroy(eval(config_str))
    ticket = KrbTicket.init_by_config(eval(config_str))
    updater = ticket.updater(interval=0.5)
    updater.start()

    def run():
        t_ticket = KrbTicket.get_by_config(eval(config_str))
        t_updater = t_ticket.updater(interval=0.5)

        assert updater == t_updater
        assert t_updater.is_alive()

        t_updater.start()
        assert ticket == t_ticket
        assert_ticket(ticket, t_ticket)

        time.sleep(3)


    executor = ThreadPoolExecutor(max_workers=10)
    for future in [executor.submit(run) for i in range(10)]:
        future.result()


@pytest.mark.parametrize('config_str', [
    'default_config(updater_class=SimpleKrbTicketUpdater)',
    'default_config(updater_class=MultiProcessKrbTicketUpdater)',
    'default_config(updater_class=SingleProcessKrbTicketUpdater)'
])
def test_skip_subsequent_updater_start_with_multithreading(config_str, caplog):
    KrbCommand.kdestroy(eval(config_str))
    KrbTicket.init_by_config(eval(config_str))

    # Subsequent updater.start should be skipped.
    executor = ThreadPoolExecutor(max_workers=5)
    for future in [executor.submit(_updater_run, eval(config_str)) for i in range(10)]:
        future.result()


@pytest.mark.parametrize('config_str', [
    'default_config(updater_class=SimpleKrbTicketUpdater)',
    'default_config(updater_class=MultiProcessKrbTicketUpdater)',
    'default_config(updater_class=SingleProcessKrbTicketUpdater)'
])
def test_skip_subsequent_updater_start_with_multiprocessing(config_str, caplog):
    KrbCommand.kdestroy(eval(config_str))
    KrbTicket.init_by_config(eval(config_str))
    # Subsequent updater.start should be skipped.
    executor = ProcessPoolExecutor(max_workers=5)
    for future in [executor.submit(_updater_run, eval(config_str)) for i in range(10)]:
        future.result()


def _updater_run(config):
    ticket = KrbTicket.init_by_config(config)
    updater = ticket.updater(interval=0.5)
    updater.start()
    updater.start()
